#######################################################################################################
# Copyright Orchid Systems 2019
# =====================================================================================================
#
# Description: A double-column data entry screen.
#              Based on the Extender Custom Table Editor UI.
#              This screen extends the functionality to include validation and labels for fields that
#              have finder tables.
#              Now has filters on finders.
#
#######################################################################################################

#######################################################################################################
# Disclaimer:
# This code is to be used with Orchid Systems' Extender module only.
# Orchid Systems provides the code on an as is basis having been developed and tested in-house 
# using a standard Sage 300 implementation with Sage sample data.
# If this code is deployed in a live production environment it is the responsibility of the End User
# and the Sage Business Partner to ensure that is is operating as required.
#######################################################################################################

from accpac import *

#Change the custom table name as required
myCustomTableName = "ADDINFO"

# This class is used 
class MyFec:
	def __init__(self, ui, fec):
		self.ui = ui
		self.fec = fec
		fec.onClick = self.click

	def click(self, btnType):
		if btnType == 4:
			finder = Finder()

			ftable = self.fec.ftable
			if ftable.find(".") != -1:
				vVICUSTH = openView("VI0011", 0)
				vVICUSTH.order(0) # TABLENAME
				vVICUSTH.put("TABLENAME", ftable)
				if vVICUSTH.read() == 0:
					ftable = vVICUSTH.get("VIEWID")
			finder.viewID = ftable

			v = openView(self.fec.ftable, 99)
			if v == None:
				showMessageBox("Could not open '" + self.fec.ftable + "'")
				return
			if v.fieldByName(self.fec.ffield) == None:
				showMessageBox("Field '" + self.fec.ffield + "' does not exist in table '" + self.fec.ftable + "'")
				return
			
			# Set the filter on the finder.
			finder.filter = self.fec.ffilter
			
			# Set the field(s) for which the finder will return a value:
			finder.returnFields = str(v.fieldByName(self.fec.ffield).index)
			
			# Associate the onOK and onCancel events with functions:
			finder.onOK = self.finder_ok
			finder.onCancel = self.finder_cancel
			
			# Display the finder:
			finder.show(self.ui)

	def finder_ok(self, e):
		# After the clicking SELECT on the finder:
		f = self.ui.ds.fieldByName(self.fec.fieldName)
		if ((f.type == FT_INT) or (f.type == FT_LONG)):
			self.ui.ds.put(self.fec.fieldName, int(e))
		else:
			self.ui.ds.put(self.fec.fieldName, e)
		ui_command("return", "4")

	def finder_cancel(self):
		ui_command("return", "1")
		
	def buildFilterString(self, raw_ffilter):
		ffilter = raw_ffilter
		
		# remove leading and trailing quotes if any have been entered.
		if ffilter[0] == "\"":
			ffilter = ffilter[1:]
			
		if ffilter[-1] == "\"":
			ffilter = ffilter[:-1]
			
		n = self.ui.ds.fields()
		for x in range(0, n):
			fld = self.ui.ds.fieldByPosition(x)
			# get the field value and replace the field reference in the finder string.
			ffilter = ffilter.replace("{" + fld.name.upper() + "}", "\"" + str(self.ui.ds.get(fld.name)) + "\"")

		return ffilter
		
class MyUI(UI):

	top_offset = 50
	#defTable = "ADDINFO"	# Name of the table that we want to edit using this screen.
	defTable = myCustomTableName
	
	def __init__(self):
	
		UI.__init__(self)
		self.vicustd = None
		self.createScreen()

	def createScreen(self):
		self.nid = 0
		self.ds = None

		# Assign to the screen close process the same process as defined for the Close button.
		self.onClose = self.onBtnClose_Click
		
		self.keyFields = []
		v = openView("VI0011") # VICUSTH
		if MyUI.defTable != None:
			self.cbTable = None
			v.put("TABLENAME", MyUI.defTable)
			if v.read() != 0:
				self.closeUI()
				return
			desc = v.get("DESC")
			if desc == "":
				desc = v.get("TABLENAME")
			self.title = "Maintain: " + desc

			self.loadTable(MyUI.defTable)
			
			# Override the datasource onAfterOpen event with the custom ds_onAfterOpen() function:
			self.ds.onAfterOpen = self.ds_onAfterOpen

		self.addButtons()
		# Show the screen:
		self.show()

	def ds_onAfterOpen(self):
		self.ds.fireEvents = False
		b = False
		for n in range(1, 10):
			f = "KEY" + str(n)
			f = getProgramKey(f, "").strip()
			if f != "":
				b = True
				self.ds.put(self.ds.fieldByPosition(n-1).name, f)
		
		self.ds.fireEvents = True
		if b == True:
			self.ds.read()
			self.ds.refreshFields()

	def loadTable(self, tableName):
		parent = self
		
		# Using the table name, open a datasource instance and apply it to the UI.
		self.ds = self.openDataSource(tableName)
		self.ds.onBefore = self.ds_onBefore
		self.ds.onBeforePut = self.ds_onBeforePut
		self.ds.onAfter = self.ds_onAfter
		self.nid += 1

		# Open the table's key definition record in VICUSTK.
		v = openView("VI0013") # VICUSTK
		v.order(1) # TABLENAME, KEYNUM
		v.put("TABLENAME", tableName)
		v.put("KEYNUM", 1)
		v.read()
		self.keyFields = []
		for n in range(1, 10):
			f = "FIELD" + str(n)
			f = v.get(f).strip()
			if f != "":
				self.keyFields.append(f.upper())	# Add to the list of key fields that comprise Key 1.

		# Open the table's field definition record in VICUSTD.
		v = openView("VI0012") # VICUSTD
		v.browse("TABLENAME=" + tableName)
		top = 165	# Set the top of the form.
			
		# Add a field-Edit control for each field defined in the table we want to edit:
		iFldCnt = 0
		prevFld = None
		while v.fetch() == 0:
			# If this is the first key field, add an extra control to act like a record navigator.
			if (v.get("FIELD").upper() == self.keyFields[0]):
				f = self.addUIField("x" + str(self.nid) + v.get("FIELD") + "F")
				f.fieldName = v.get("FIELD")
			
				f.dataSource = self.ds
				f.width = 2000
				f.top = top
				top += f.height + 75
				f.left = f.width + f.labelWidth + 720

				f.onClick = self.fecKeyField_click
				f.hasFinder = True
				f.hasArrows = True
				f.hasNew = True
				f.labelWidth = 0
		
			iFldCnt += 1
			
			
			if iFldCnt % 2 == 0:
				f = parent.addUIField("x" + str(self.nid) + v.get("FIELD"))
				f.fieldName = v.get("FIELD")
				f.dataSource = self.ds
				f.width = 5000
				f.top = prevFld.top
				f.left = prevFld.width + 500
				f.labelWidth = 120
			else:
				# Continue as normal.
				f = parent.addUIField("x" + str(self.nid) + v.get("FIELD"))
				f.fieldName = v.get("FIELD")
				f.dataSource = self.ds
				f.width = 5000
				f.top = top
				top += f.height + 75
				f.labelWidth = 120
			
			prevFld = f
			
			if v.get("DATATYPE") == 1: # Text
				if v.get("SIZE") >= 30:
					f.width = 7000
			elif v.get("DATATYPE") == 9: # bool (displays as a checkbox)
				f.borderStyle = 0
				f.labelWidth = 0
			
			if (v.get("FTABLE").strip() != "") and (v.get("FFIELD").strip() != ""):
				# A finder table and finder field have been defined for this field:
				f.hasFinder = True
				
				# create a new class instance that we can associate to this field-edit control.
				# This is so that we can identify a field-edit control to a unique finder.
				# Otherwise we don't know which fec had its finder triggered.
				m = MyFec(self, f)
				
				f.ftable = v.get("FTABLE").strip()
				f.ffield = v.get("FFIELD").strip()
				f.ffilter = v.get("FFILTER").strip()
				
				# Add labels to those fields that have finders associated with them.
				l = self.addLabel("lbl" + f.id)
				l.width = 3000
				l.top = f.top + MyUI.top_offset
				l.left = f.width + 500

	def getFilter(self, commentStr):
		# strip out the filter.
		commSplit = commentStr.split(";")
		for c in commSplit:
			c = c.upper()
			if (c.find("=") != -1):
				c2 = c.split("=", 1)
				if (c2[0].strip() == "FILTER"):
					return c2[1].strip()
					
		return ""

	def getLabelDescField(self, commentStr):
		# strip out the filter.
		commSplit = commentStr.split(";")
		for c in commSplit:
			c = c.upper()
			if (c.find("=") != -1):
				c2 = c.split("=", 1)
				if (c2[0].strip() == "DESC"):
					return c2[1].strip() # We are expecting only 1 x description field.
					
		return ""
		
	def addButtons(self):
		# In the event they are already attached to the UI, remove the buttons.
		self.removeControl("btnSave")
		self.removeControl("btnDelete")
		self.removeControl("btnClose")

		# Now add the buttons:
		btn = self.addButton("btnSave", "&Add")
		btn.top = -150 - btn.height
		btn.width = 1065
		btn.onClick = self.onBtnSave_Click
		btn.disable()
		self.btnSave = btn

		btn = self.addButton("btnDelete", "&Delete...")
		btn.top = self.btnSave.top
		btn.width = 1065
		btn.left = self.btnSave.left + self.btnSave.width + 80
		btn.onClick = self.onBtnDelete_Click
		btn.disable()
		self.btnDelete = btn

		btn = self.addButton("btnClose", "&Close")
		btn.top = self.btnSave.top
		btn.width = 1065
		btn.left = -btn.width - 150
		btn.onClick = self.onBtnClose_Click
		self.btnClose = btn

		
###### UI Template Functions ######################################
# 
# Do not modify these functions unless you know what you're doing.
#
	def onBtnClose_Click(self):
		self.CheckSaveRecord(self.onBtnClose_Click_part2)

	def onBtnClose_Click_part2(self, bProceed):
		if bProceed == True:
			self.closeUI()
		else:
			self.abortClose()

	def onBtnSave_Click(self):
		self.saveRecord()

	def onBtnDelete_Click(self):
		self.message("Are you sure you want to delete this record?", "Delete?", "QUESTION", "YESNO", self.onBtnDelete_confirmed)

	def onBtnDelete_confirmed(self, r):
		if r == "YES":
			self.ds.fireEvents = False
			self.ds.delete()
			self.ds.fireEvents = True
			self.NewHeaderLoaded()

	def isKeyField(self, fieldName):
		fieldName = fieldName.upper()
		for k in self.keyFields:
			if fieldName == k:
				return True
		return False

	def ds_onBeforePut(self, fieldName, newValue):
		if self.isKeyField(fieldName):
			# amend the check to only happen on the first segment key field.
			if (fieldName == self.keyFields[0]):
				self.CheckSaveRecord()
		# Validate the field value if it's a field with an associated finder table.
		return self.validateField(fieldName, newValue)

	def ds_onBefore(self, e, fieldName, value):
		if e == "FETCH" or e.startswith("GO") or e == "INIT":
			self.CheckSaveRecord()
		elif e == "DELETE":
			self.message("Confirm delete", "Delete?", "QUESTION", "YESNO", self.onConfirmDelete)

	def onConfirmDelete(self, r):
		if r != "YES":
			ui_command("return", "ABORT")

	# Modified to cater for labels.
	def ds_onAfter(self, e, fieldName, value):
		if e == "FETCH" or e.startswith("GO") or e == "INIT" or e == "READ" or e == "DELETE" or e == "INSERT" or e == "UPDATE" or e == "POST":
			self.NewHeaderLoaded()
		elif e == "PUT":
			if self.isKeyField(fieldName):
				self.OnDocumentNumberChanged()
			
			# New code: Update the label on AFTER a viewPut() call.
			self.updateDescLabel(fieldName)

	def OnDocumentNumberChanged(self):
		if self.ds.exists() == 1:
			self.ds.read()
			self.ds.refreshFields()
		else:
			self.ds.fireEvents = False
			values = []
			for k in self.keyFields:
				values.append(self.ds.get(k))
			self.ds.order(0)
			self.ds.recordGenerate()
			n = 0
			for v in values:
				if isinstance(v, str) == False:
					self.ds.put(self.keyFields[n], v)
				elif v.strip() != "":
					self.ds.put(self.keyFields[n], v)
				n += 1
			self.ds.fireEvents = True
		self.NewHeaderLoaded()

	def CheckSaveRecord(self, cb=None):
		if self.UnsavedChangesExist() == False:
			if cb != None:
				cb(True)
			return
		if cb == None:
			self.csrCb = self.csr_default
		else:
			self.csrCb = cb
		self.message("Save changes?", "Record has been modified", "QUESTION", "YESNOCANCEL", self.onConfirmSave)

	def csr_default(self, bProceed):
		if bProceed == False:
			ui_command("return", "ABORT")

	def saveRecord(self):
		if self.ds.exists() == 0:
			return self.ds.insert()
		else:
			return self.ds.update()

	def onConfirmSave(self, r):
		if r == "YES":
			if self.saveRecord() != 0:
				self.csrCb(False)
			else:
				self.csrCb(True)
		elif r == "CANCEL":
			self.csrCb(False)
		else:
			self.csrCb(True)

	def UnsavedChangesExist(self):
		if self.ds == None:
			return False
		if self.ds.dirty() != 0:
			return True
		return False

	def NewHeaderLoaded(self):
		self.UpdateButtons()
		self.updateDescLabel()

	def UpdateButtons(self):
		if self.ds.exists() == 0:
			self.btnSave.setCaption("&Add")
			self.btnDelete.disable()
		else:
			self.btnSave.setCaption("&Save")
			self.btnDelete.enable()

		for k in self.keyFields:
			if str(self.ds.get(k)).strip() == "":
				self.btnSave.disable()
				return
		self.btnSave.enable()

	def fecKeyField_click(self, btnType):
		if btnType == 4:
			finder = Finder()
			finder.viewID = self.ds.rotoID
			finder.onOK = self.name_finder_ok
			finder.onCancel = self.name_finder_cancel
			if len(self.keyFields) > 1:
				finder.returnFields = ""
				for n in range(1, len(self.keyFields)+1):
					finder.returnFields = finder.returnFields + str(n) + ","
				finder.returnFields = finder.returnFields[0:-1]
			# TODO - what fields to show
			finder.show(self)

	def name_finder_ok(self, e):
		self.finderValue = e
		self.CheckSaveRecord(self.name_finder_ok_part2)

	def name_finder_ok_part2(self, bProceed):
		if bProceed == False:
			ui_command("return", "1")
			return
		
		# Turn off datasource events to avoid having them trigger on the following calls:
		self.ds.fireEvents = False
		if len(self.keyFields) > 1:
			for n in range(0, len(self.keyFields)):
				f = self.ds.fieldByName(self.keyFields[n])
				if ((f.type == FT_INT) or (f.type == FT_LONG)):
					self.ds.put(self.keyFields[n], int(self.finderValue[n]))
				elif f.type == FT_DATE:
					# value will be YYYYMMDDhhnnss00
					self.ds.put(self.keyFields[n], self.finderValue[n][0:8])
				elif f.type == FT_TIME:
					# value will be YYYYMMDDhhnnss00
					self.ds.put(self.keyFields[n], self.finderValue[n][8:8+8])
				else:
					self.ds.put(self.keyFields[n], self.finderValue[n])
		else:
			f = self.ds.fieldByName(self.keyFields[0])
			if ((f.type == FT_INT) or (f.type == FT_LONG)):
				self.ds.put(self.keyFields[0], int(self.finderValue))
			else:
				self.ds.put(self.keyFields[0], self.finderValue)
				
		# Turn the datasource events back on:
		self.ds.fireEvents = True
		self.ds.read()
		ui_command("return", "4")

	def name_finder_cancel(self):
		ui_command("return", "1")


	########################### Custom Routines ##############################
	
	########################### Field Validation #############################
	def validateField(self, fieldName, newValue):
		rvspyTrace("validateField")
		
		if str(newValue).strip() == "":
			return True
		
		fieldName = fieldName.strip()
		f = self.getFieldControlByName(fieldName)
		if f != None:
			if f.hasFinder == True:
				# we can only automatically validate fields that have finders.
				# this is because we are assuming the finder table points to the field's master table.
				# For example, an item field will have a finder pointing to IC0310, the item record table.
				# We can use this to find out if a field value exists in the finder table.
				ffield, ftable, fdesc, fdescField = self.getFinderTableField(fieldName)
				if ftable != "":
					v = openView(ftable)
					v.recordClear()
					sFilter = ffield + "=\"" + str(newValue).strip() + "\""
					v.browse(sFilter, 1)
					if v.fetch() == 0:
						# record exists with this key value.
						v.close()
						return True
					else:
						sErrMsg = "Invalid Entry. " + fdesc + " [" + newValue + "] not found in the database."
						showMessageBox(sErrMsg)
						v.close()
						return False
				else:
					# Finder table not found. Possibly an error in the setup?
					return True	# Cannot validate the value in newValue.
			else:
				# Cannot validate. This field has no associated finder table.
				return True
		else:
			# Not a UI control field that we can validate.
			return True
		
	########################### Label Updates ################################
	def updateDescLabel(self, fname=""):
		rvspyTrace("updateLabel")
		
		if (fname == ""):
			# Update the labels for all field-edit controls that have finders.
			# Note: the record navigation control has a finder, but it's not defined in the table definition, so there is no label to update.
			for f in self.controls:
				if f.type == "EDIT":
					if f.hasFinder == True:
						descValue = ""
						fname = f.fieldName
						# From the table definition record, get information on the finder table, 
						# field, the description field, and the field description.
						ffield, ftable, fdesc, fdescField = self.getFinderTableField(f.fieldName)
						if ftable != "":
							# find a label for it.
							v = openView(ftable)
							v.recordClear()
							fvalue = self.ds.get(fname)
							rvspyTrace(fname + " Value = ["+str(fvalue)+"]")
							if fvalue != None:
								if str(fvalue) != "":
									v.put(ffield, fvalue)
									if v.read() == 0:
										# read description field. Get it from the table definition.
										if fdescField != "":
											descValue = str(v.get(fdescField)).strip()
											# apply it to the label. Use the setText method.
											rvspyTrace("label: " + descValue)
											lblID = "lbl" + f.id
											l = self.getLabelControlByID(lblID)
											if l != None:
												l.setText(descValue)
									else:
										# try browsing for the record:
										v.recordClear()
										sFilter = ffield + " = \"" + str(fvalue) + "\""
										v.browse(sFilter, 1)
										if v.fetch() == 0:
											# read description field. Get it from the table definition.
											if fdescField != "":
												descValue = str(v.get(fdescField)).strip()
												# apply it to the label. Use the setText method.
												lblID = "lbl" + f.id
												l = self.getLabelControlByID(lblID)
												if l != None:
													l.setText(descValue)
								else:
									self.applyBlankLabel(f.id, fdescField)
							else:
								# The field value in the dataSource is None.
								self.applyBlankLabel(f.id, fdescField)
								
							v.close()
			
		else:
			# Update the label of a specific field.
			f = self.getFieldControlByName(fname)
			if f != None:
				if f.hasFinder == True:
					descValue = ""
					ffield, ftable, fdesc, fdescField = self.getFinderTableField(f.fieldName)
					rvspyTrace(ffield + ", " + ftable + ", " + fdesc)
					if ftable != "":
						# find a label for it.
						v = openView(ftable)
						v.recordClear()
						fvalue = self.ds.get(fname)
						rvspyTrace(fname + " Value = ["+str(fvalue)+"]")
						if fvalue != None:
							if fvalue != "":
								v.put(ffield, fvalue)
								if v.read() == 0:
									rvspyTrace("Master Record found.")
									# read description field. Get it from the table definition.
									if fdescField != "":
										descValue = str(v.get(fdescField)).strip()
										# apply it to the label. Use the setText method.
										rvspyTrace("label: " + descValue)
										lblID = "lbl" + f.id
										l = self.getLabelControlByID(lblID)
										if l != None:
											l.setText(descValue)
								else:
									# try browsing for the record:
									v.recordClear()
									sFilter = ffield + " = \"" + str(fvalue) + "\""
									v.browse(sFilter, 1)
									if v.fetch() == 0:
										# read description field. Get it from the table definition.
										if fdescField != "":
											descValue = str(v.get(fdescField)).strip()
											# apply it to the label. Use the setText method.
											lblID = "lbl" + f.id
											l = self.getLabelControlByID(lblID)
											if l != None:
												l.setText(descValue)

							else:
								# The field is blank.
								self.applyBlankLabel(f.id, fdescField)
						else:
							# The field is empty.
							self.applyBlankLabel(f.id, fdescField)
							
						v.close()

	def applyBlankLabel(self, fldID, fdescField):
		# used when initializing a new record.
		rvspyTrace("applyBlankLabel")
		
		descValue = ""
		if fdescField != "":
			# build up the label ID string. Use the UIField ID.
			lblID = "lbl" + fldID
			l = self.getLabelControlByID(lblID)
			if l != None:
				# apply it to the label. Use the setText method.
				l.setText(descValue)
						
	def getFinderTableField(self, fname):
		rvspyTrace("getFinderTableField")

		ffield = ""
		ftable = ""
		fdesc = ""
		fdescField = ""
		
		if self.vicustd == None:
			self.vicustd = openView("VI0012") # VICUSTD
		
		sFilter = "TABLENAME = \"" + MyUI.defTable + "\" AND FIELD = \"" + fname + "\""
		rvspyTrace(sFilter)
		self.vicustd.recordClear()
		self.vicustd.browse(sFilter, 1)
		while self.vicustd.fetch() == 0:
			ffield = self.vicustd.get("FFIELD").strip()
			ftable = self.vicustd.get("FTABLE").strip()
			fdesc = self.vicustd.get("DESC").strip()
			commentStr = self.vicustd.get("COMMENTS").strip()
			
			fdescField = self.getLabelDescField(commentStr)

		return ffield, ftable, fdesc, fdescField
		
	def getLabelControlByID(self, lblID):
		rvspyTrace("getLabelControlByID")
		for f in self.controls:
			if f.type == "LABEL":
				if f.id == lblID:
					# This is the label control we want.
					return f
					
		return None
		
	def getFieldControlByName(self, fname):
		rvspyTrace("getFieldControlByName")
		for f in self.controls:
			if f.type == "EDIT":
				if f.fieldName.strip() == fname.strip():
					# This is the UI control we want.
					rvspyTrace("Found the right control.")
					return f
					
		return None
	
########################## MAIN ##########################################
def main(args):
	# Call the UI Class:
	MyUI()
